from typing import Optional, TypeVar, Generic
from pydantic import BaseModel

T = TypeVar('T', bound=BaseModel)


class AbstractRepository(Generic[T]):
    def __init__(self, database):
        super().__init__()
        self.__database = database
        self.__document_class = self.__orig_bases__[0].__args__[0]
        self.__collection_name = self.Meta.collection_name
        self.__validate()

    def get_collection(self):
        return self.__database[self.__collection_name]

    def __validate(self):
        if not issubclass(self.__document_class, BaseModel):
            raise Exception('Document class should inherit BaseModel')
        if 'id' not in self.__document_class.__fields__:
            raise Exception('Document class should have id field')
        if not self.__collection_name:
            raise Exception('Meta should contain collection name')

    def to_document(self, model: T) -> dict:
        result = model.dict()
        result.pop('id')
        if model.id:
            result['_id'] = model.id
        return result

    def __to_query(self, data: dict):
        query = data.copy()
        if 'id' in data:
            query['_id'] = query.pop('id')
        return query

    def to_model(self, data: dict) -> T:
        data_copy = data.copy()
        if '_id' in data_copy:
            data_copy['id'] = data_copy.pop('_id')
        return self.__document_class.parse_obj(data_copy)

    def save(self, model: T):
        document = self.to_document(model)

        if model.id:
            mongo_id = document.pop('_id')
            self.get_collection().update_one({'_id': mongo_id}, {'$set': document})
            return

        result = self.get_collection().insert_one(document)
        model.id = result.inserted_id
        return result

    def delete(self, model: T):
        return self.get_collection().delete_one({'_id': model.id})

    def find_one_by_id(self, id: str) -> Optional[T]:
        return self.find_one_by({'id': id})

    def find_one_by(self, query: dict) -> Optional[T]:
        result = self.get_collection().find_one(self.__to_query(query))
        return self.to_model(result) if result else None

    def find_by(
        self,
        query: dict,
        skip: Optional[int] = None,
        limit: Optional[int] = None,
        sort=None
    ):
        cursor = self.get_collection().find(self.__to_query(query))
        if limit:
            cursor.limit(limit)
        if skip:
            cursor.skip(skip)
        if sort:
            cursor.sort(sort)
        return map(self.to_model, cursor)



from bson import ObjectId


class ObjectIdField():
    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, value):
        if not ObjectId.is_valid(value):
            raise ValueError("Invalid id")

        return ObjectId(value)

    @classmethod
    def __modify_schema__(cls, field_schema):
        field_schema.update(type="string")



import pytest
import mongomock
from bson import ObjectId
from typing import List
from pydantic_mongo import AbstractRepository, ObjectIdField
from pydantic import BaseModel


class Foo(BaseModel):
    count: int
    size: float = None


class Bar(BaseModel):
    apple = 'x'
    banana = 'y'


class Spam(BaseModel):
    id: ObjectIdField = None
    foo: Foo
    bars: List[Bar]

    class Config:
        json_encoders = {ObjectId: str}


class SpamRepository(AbstractRepository[Spam]):
    class Meta:
        collection_name = 'spams'


@pytest.fixture
def database():
    return mongomock.MongoClient().db.collection


class TestRepository:
    def test_save(self, database):
        spam_repository = SpamRepository(database=database)
        foo = Foo(count=1, size=1.0)
        bar = Bar()
        spam = Spam(foo=foo, bars=[bar])
        spam_repository.save(spam)

        assert {
            '_id': ObjectId(spam.id),
            'foo': {'count': 1, 'size': 1.0},
            'bars': [{'apple': 'x', 'banana': 'y'}]
        } == database['spams'].find()[0]

        spam.foo.count = 2
        spam_repository.save(spam)

        assert {
            '_id': ObjectId(spam.id),
            'foo': {'count': 2, 'size': 1.0},
            'bars': [{'apple': 'x', 'banana': 'y'}]
        } == database['spams'].find()[0]

    def test_delete(self, database):
        spam_repository = SpamRepository(database=database)
        foo = Foo(count=1, size=1.0)
        bar = Bar()
        spam = Spam(foo=foo, bars=[bar])
        spam_repository.save(spam)

        result = spam_repository.find_one_by_id(spam.id)
        assert result is not None

        spam_repository.delete(spam)
        result = spam_repository.find_one_by_id(spam.id)
        assert result is None

    def test_find_by_id(self, database):
        spam_id = ObjectId('611827f2878b88b49ebb69fc')
        database.spams.insert_one({
            '_id': spam_id,
            'foo': {'count': 2, 'size': 1.0},
            'bars': [{'apple': 'x', 'banana': 'y'}]
        })

        spam_repository = SpamRepository(database=database)
        result = spam_repository.find_one_by_id(spam_id)

        assert issubclass(Spam, type(result))
        assert spam_id == result.id
        assert 'x' == result.bars[0].apple

    def test_find_by(self, database):
        database.spams.insert_many([
            {
                'foo': {'count': 2, 'size': 1.0},
                'bars': [{'apple': 'x', 'banana': 'y'}]
            },
            {
                'foo': {'count': 3, 'size': 1.0},
                'bars': [{'apple': 'x', 'banana': 'y'}]
            },
        ])

        spam_repository = SpamRepository(database=database)

        # Simple Find
        result = spam_repository.find_by({})
        results = [x for x in result]
        assert 2 == len(results)
        assert 2 == results[0].foo.count
        assert 3 == results[1].foo.count

        # Find with optional parameters
        result = spam_repository.find_by({}, skip=10, limit=10, sort=[('foo.count', 1)])
        results = [x for x in result]
        assert 0 == len(results)

    def test_invalid_model_class(self, database):
        class BrokenRepository(AbstractRepository[int]):
            class Meta:
                collection_name = 'spams'
        with pytest.raises(Exception):
            BrokenRepository(database=database)

    def test_invalid_model_id_field(self, database):
        class NoIdModel(BaseModel):
            something: str

        class BrokenRepository(AbstractRepository[NoIdModel]):
            class Meta:
                collection_name = 'spams'

        with pytest.raises(Exception):
            BrokenRepository(database=database)

    def test_invalid_model_collection_name(self, database):
        class BrokenRepository(AbstractRepository[Spam]):
            class Meta:
                collection_name = None

        with pytest.raises(Exception):
            BrokenRepository(database=database)



import pytest
from bson import ObjectId
from pydantic_mongo import ObjectIdField
from pydantic import BaseModel
from pydantic.error_wrappers import ValidationError


class User(BaseModel):
    id: ObjectIdField = None

    class Config:
        json_encoders = {ObjectId: str}


class TestFields:
    def test_object_id_validation(self):
        with pytest.raises(ValidationError):
            User.parse_obj({'id': 'lala'})
        User.parse_obj({'id': '611827f2878b88b49ebb69fc'})

    def test_object_id_serialize(self):
        lala = User(id=ObjectId('611827f2878b88b49ebb69fc'))
        json_result = lala.json()
        assert '{"id": "611827f2878b88b49ebb69fc"}' == json_result

    def test_modify_schema(self):
        user = User(id=ObjectId('611827f2878b88b49ebb69fc'))
        schema = user.schema()
        assert {
            'title': 'User',
            'type': 'object',
            'properties': {
                'id': {
                    'title': 'Id',
                    'type': 'string'
                }
            }
        } == schema



from setuptools import setup

long_description = open('README.rst', 'r').read()

setup(
    name='pydantic-mongo',
    version='0.0.1',
    packages=['pydantic_mongo'],
    setup_requires=['wheel'],
    install_requires=[
        'pymongo>=3.12,<4.0',
        'pydantic>=1.6.2,<2.0.0'
    ],
    entry_points={
        "console_scripts": [
            "pydantic_mongo = pydantic_mongo.__main__:__main__"
        ],
    },
    description="Document object mapper for pydantic and pymongo",
    long_description=long_description,
    url='https://github.com/jefersondaniel/pydantic-mongo',
    author='Jeferson Daniel',
    author_email='jeferson.daniel412@gmail.com',
    license='MIT',
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
    ],
    python_requires=">=3.6"
)
